//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  gmodapp.cpp: main application code
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "gmodapp.h"
#include "interface.h"
#include "support.h"
#include "modinforeader.h"
#include "modconfmgr.h"
#include "moduletypes.h"
#include "modgui.h"

#include <iostream>
using std::cout;
using std::endl;

#include <cstdio>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/param.h>

using namespace gmodcfg;

extern Application* TheApp;

enum { MODULE_NAME_COLUMN, SHORT_DESC_COLUMN, NUM_COLUMNS };

//--------------------------------------------------------------------------//
//  C'tor and d'tor
//--------------------------------------------------------------------------//
Application::Application( const string& modPath, const string& modConfFile,
                          const string& lang, int v )
            : TopWidget(0), TreeView(0), StatusBar(0),
              ModulesPath(modPath), ModulesConfFile(modConfFile),
              Locale(lang), Verbosity(v), ModGuiMgr(0)
{
    ModConfMgr = new ModulesConfManager( modConfFile, Verbosity );
    if ( Verbosity )
        cout << "Modules dir: " << modPath << endl;

    ModInfoReader = new ModuleInfoReader( Verbosity );
}

Application::~Application()
{
    delete ModInfoReader;
    delete ModConfMgr;
    delete ModGuiMgr;
}

//--------------------------------------------------------------------------//
//  Store the current module's settings into the ModulesConfManager
//--------------------------------------------------------------------------//
void
Application::storeCurrentModuleSettings()
{
    const Module* currMod = ModGuiMgr->GetCurrentModule();
    if ( !currMod )
        return;

    const string& currModName = currMod->GetName();
    string options = ModGuiMgr->GenerateOptionsString();
    ModConfMgr->SetOptions( currModName, options );
    if ( Verbosity )
        cout << "Setting " << currModName << "'s options to \""
             << options << "\"" << endl;
}

//--------------------------------------------------------------------------//
//  Helper function: excludes . and .. from our directory search (scandir)
//--------------------------------------------------------------------------//
static int
filterFiles( const struct direct* entry )
{
    return ! ( (strcmp(entry->d_name, ".") == 0)
            || (strcmp(entry->d_name, "..") == 0) );
}

//--------------------------------------------------------------------------//
//  At a given directory level in the modules hierarchy, create an entry in
//  the tree_store and recursively call ourself on the files/dirs it contains.
//--------------------------------------------------------------------------//
static void
populateTreestore( GtkTreeStore* treestore, GtkTreeIter* parentIter,
                    const string& pathName, bool topLevel = false )
{
    struct direct** namelist;
    int count = scandir( pathName.c_str(), &namelist, filterFiles, 0 );
    const char* itemName = basename( pathName.c_str() );
 
    GtkTreeIter iter;

    // Start with only the subdirs below pathName when at the top
    if ( !topLevel )
    {
        gtk_tree_store_append( treestore, &iter, parentIter );

        // If no files found: this is a leaf (module)
        if ( count <= 0 )
        {
            // TODO: display module's short description in 2nd column
            gtk_tree_store_set( treestore, &iter, MODULE_NAME_COLUMN, itemName,
                                SHORT_DESC_COLUMN, "", -1 );
            return;
        }

        // That's a directory
        gtk_tree_store_set( treestore, &iter, MODULE_NAME_COLUMN, itemName,
                            SHORT_DESC_COLUMN, "", -1 );
    }

    for( int i=0; i<count; ++i )
    {
        const char* childName = namelist[i]->d_name;
        string childPath = pathName + "/" + childName;

        // go further down, in case this is a dir
        populateTreestore( treestore, ( topLevel ? 0 : &iter ), childPath );
    }
}

//--------------------------------------------------------------------------//
//  Create 2 columns for our treeview
//--------------------------------------------------------------------------//
static void
addColumns( GtkTreeView* treeview, GtkTreeStore* treestore )
{
    GtkTreeModel* model = GTK_TREE_MODEL(treestore);
    gtk_tree_view_set_model( treeview, model );

    // 1st column: module names
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    g_object_set( G_OBJECT(renderer), "xalign", 0.0, NULL );
    gint col_offset = gtk_tree_view_insert_column_with_attributes(
                            GTK_TREE_VIEW(treeview), -1, "Module",
                            renderer, "text", MODULE_NAME_COLUMN, NULL );
    GtkTreeViewColumn* column = gtk_tree_view_get_column(
                                GTK_TREE_VIEW(treeview), col_offset - 1 );
    gtk_tree_view_column_set_clickable( GTK_TREE_VIEW_COLUMN(column), TRUE );

    // 2nd column: short description
    renderer = gtk_cell_renderer_text_new();
    g_object_set( G_OBJECT(renderer), "xalign", 0.0, NULL );
    g_object_set_data( G_OBJECT(renderer), "column", (gint*)SHORT_DESC_COLUMN );
    col_offset = gtk_tree_view_insert_column_with_attributes(
                            GTK_TREE_VIEW(treeview), -1, "Short Description",
                            renderer, "text", SHORT_DESC_COLUMN, NULL );
    column = gtk_tree_view_get_column( GTK_TREE_VIEW(treeview),
                                        col_offset - 1 );
    gtk_tree_view_column_set_clickable( GTK_TREE_VIEW_COLUMN(column), TRUE );
}

//--------------------------------------------------------------------------//
//  Fill in the treeview with the module hierarchy
//--------------------------------------------------------------------------//
static void
populateTreeview( GtkTreeView* treeView, const string& modulesPath )
{
    GtkTreeStore* treestore = gtk_tree_store_new( NUM_COLUMNS,
                                                G_TYPE_STRING, G_TYPE_STRING );
    populateTreestore( treestore, NULL, modulesPath, true );
    addColumns( treeView, treestore );
}

//--------------------------------------------------------------------------//
//  Treeview's "changed" callback
//--------------------------------------------------------------------------//
static void
selectionChangedCallback( GtkTreeSelection* select, gpointer data )
{
    GtkTreeModel* model = (GtkTreeModel*)data;
    GtkTreeIter iter;

    if ( gtk_tree_selection_get_selected( select, &model, &iter ) )
    {
        char* itemName;
        gtk_tree_model_get( model, &iter, MODULE_NAME_COLUMN, &itemName, -1 );

        // if it's a module name update the notebook GUI
        string str( itemName );
        int dot = str.find_last_of(".");
        if ( dot == string::npos || str.substr( dot, 2 ) != ".o" )
            return;

        // strip the suffix off of the module name
        str.erase( dot, 2 );

        // read module's information and update the GUI
        const Module* mod = TheApp->GetModuleInfo( str );
        TheApp->BuildModuleGui( mod );
    }
}

//--------------------------------------------------------------------------//
//  Install a callback to detect click on the treeview
//--------------------------------------------------------------------------//
static void
installTreeviewCallback( GtkTreeView* treeView )
{
    GtkTreeModel* model = gtk_tree_view_get_model( treeView );
    GtkTreeSelection* select = gtk_tree_view_get_selection( treeView );

    g_signal_connect( G_OBJECT( select ), "changed",
                      G_CALLBACK( selectionChangedCallback ), model );
}

//--------------------------------------------------------------------------//
//  Application's "delete_event" callback
//--------------------------------------------------------------------------//
static void
deleteEventCallback( GtkWidget *widget, GdkEvent *event, gpointer data )
{
    TheApp->Quit();
}

//--------------------------------------------------------------------------//
//  Builds the static GUI and enters the GTK event loop
//--------------------------------------------------------------------------//
void
Application::Start()
{
    TopWidget = create_app();
    gtk_widget_show( TopWidget );
    StatusBar = GNOME_APPBAR( lookup_widget( TopWidget, "appbar" ) );

    // Read the contents of the modules.conf file
    if ( !ModConfMgr->ReadModulesConf() )
    {
        gchar* msg = g_strdup_printf(
             _("Could not read the contents of %s"), ModulesConfFile.c_str() );
        gnome_appbar_set_status( StatusBar, msg );
        g_free( msg );
    }

    ModGuiMgr = new ModuleGuiBuilder( TopWidget, Verbosity );

    // delete_event happens when the user closes the window
    gtk_signal_connect( GTK_OBJECT( TopWidget ), "delete_event",
                        GTK_SIGNAL_FUNC( deleteEventCallback ), NULL );

    // Populate the tree view
    TreeView = GTK_TREE_VIEW( lookup_widget( TopWidget, "treeview" ) );
    populateTreeview( TreeView, ModulesPath );
    installTreeviewCallback( TreeView );


    gtk_main();
}

//--------------------------------------------------------------------------//
//  Pop up a dialog if some changes are not yet saved
//--------------------------------------------------------------------------//
void
Application::Quit()
{
    storeCurrentModuleSettings();

    if ( ModConfMgr->HasUnsavedChanges() )
    {
        GtkWidget* quit = create_quitdialog();
        gtk_window_set_transient_for( GTK_WINDOW(quit), GTK_WINDOW(TopWidget) );
        gtk_window_present( GTK_WINDOW(quit) );
    }
    else
        gtk_main_quit();
}

//--------------------------------------------------------------------------//
//  Retrieve module information and return a new Module object
//--------------------------------------------------------------------------//
const Module*
Application::GetModuleInfo( const string& str ) const
{
    return ModInfoReader->GetInfo( str, GetLocale() );
}

//--------------------------------------------------------------------------//
//  Dynamically fill in the notebook tabs with Module data
//--------------------------------------------------------------------------//
void
Application::BuildModuleGui( const Module* mod )
{
    if ( !mod )
        return;

    // Save current module's settings
    storeCurrentModuleSettings();

    // Then move on with the new module
    ModGuiMgr->BuildGui( mod, ModConfMgr->GetOptions( mod->GetName() ) );
}

//--------------------------------------------------------------------------//
//  Discard the changes made so far to the current module's parameters
//--------------------------------------------------------------------------//
void
Application::DiscardChanges()
{
    const Module* currMod = ModGuiMgr->GetCurrentModule();
    if ( !currMod )
        return;

    const string& currModName = currMod->GetName();
    if ( Verbosity )
        cout << "Discarding changes for module " << currModName << endl;

    // Forget the changes and rebuild the GUI, to restore the widgets states
    ModConfMgr->SetOptions( currModName, "" );
    GtkNotebook* notebook = GTK_NOTEBOOK(lookup_widget(TopWidget, "notebook"));
    ModGuiMgr->BuildGui( currMod, ModConfMgr->GetOptions( currModName ) );
}

//--------------------------------------------------------------------------//
//  Save all the options changes into the modules.conf file
//--------------------------------------------------------------------------//
void
Application::SaveAllOptions()
{
    storeCurrentModuleSettings();
    if ( !ModConfMgr->WriteAllOptions() )
    {
        gchar* msg = g_strdup_printf(
                _("Could not save changes to %s, is the file writable?"),
                ModulesConfFile.c_str() );
        gnome_appbar_set_status( StatusBar, msg );
        g_free( msg );
    }
}

//--------------------------------------------------------------------------//
//  Load (or reload) the current module with the current (modified) options
//--------------------------------------------------------------------------//
void
Application::ApplyOptions() const
{
    const Module* currMod = ModGuiMgr->GetCurrentModule();
    if ( !currMod )
        return;

    // TODO
    // Check with lsmod if module's there, run rmmod, check lsmod again
    // and run modprobe
}

//--------------------------------------------------------------------------//
//  Show the about box
//--------------------------------------------------------------------------//
void
Application::ShowAboutBox() const
{
    GtkWidget* about = create_about();
    gtk_window_set_transient_for( GTK_WINDOW(about), GTK_WINDOW(TopWidget) );
    gtk_window_present( GTK_WINDOW(about) );
}

