//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modgui.cpp: build GUI of a module
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <iterator>
#include <sstream>
#include <iostream>

using std::cout;
using std::endl;

#include "modgui.h"
#include "moduletypes.h"
#include "support.h"
#include "xsatypes.h"
#include "xsagui.h"

using namespace gmodcfg;

//--------------------------------------------------------------------------//
//  Create a left justified label in a table
//--------------------------------------------------------------------------//
static GtkWidget*
makeLabel( GtkTable* table, const string& str, int row, int col )
{
    GtkWidget* label = gtk_label_new( str.c_str() );
    gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
    gtk_label_set_line_wrap( GTK_LABEL(label), TRUE );
    gtk_misc_set_alignment( GTK_MISC(label), 0, 0.5 );
    gtk_widget_show( label );
    gtk_table_attach( table, label, col, col+1, row, row+1,
                      (GtkAttachOptions)(GTK_FILL),
                      (GtkAttachOptions)(GTK_FILL), 1 , 1 );
    return label;
}

//--------------------------------------------------------------------------//
//  Create a window with automatic scrollbars
//--------------------------------------------------------------------------//
static GtkWidget*
makeScrolledWindow()
{
    GtkWidget* scroll = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(scroll),
                                    GTK_POLICY_AUTOMATIC,
                                    GTK_POLICY_AUTOMATIC );
    gtk_widget_show( scroll );
    return scroll;
}

//--------------------------------------------------------------------------//
//  Create a table inside a scrolled window
//--------------------------------------------------------------------------//
static GtkTable*
makeTable( GtkWidget* scroll, int numRows, int numCols )
{
    GtkWidget* table = gtk_table_new( numRows, numCols, FALSE );
    gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(scroll), table );
    gtk_widget_show( table );
    return GTK_TABLE( table );
}

//--------------------------------------------------------------------------//
//  Add a widget to a table cell, and align it on the left
//--------------------------------------------------------------------------//
static void
addToTable( GtkTable* table, GtkWidget* widget, int row, int col )
{
    GtkWidget* align = gtk_alignment_new( 0, 0.5, 0, 0 );
    gtk_container_add( GTK_CONTAINER(align), widget );
    gtk_widget_show( align );
    gtk_table_attach( table, align, col, col+1, row, row+1,
                      (GtkAttachOptions)(GTK_FILL),
                      (GtkAttachOptions)(GTK_FILL), 1, 1 );
}

//--------------------------------------------------------------------------//
//  Replace a tab's content by a new widget 
//--------------------------------------------------------------------------//
static void
replaceTab( GtkNotebook* notebook, gint tabNum, GtkWidget* newTab )
{
    GtkWidget* oldpage = gtk_notebook_get_nth_page( notebook, tabNum );
    GtkWidget* label = gtk_notebook_get_tab_label( notebook, oldpage );
    gtk_notebook_remove_page( notebook, tabNum );
    gtk_notebook_insert_page( notebook, newTab, label, tabNum );
}

//--------------------------------------------------------------------------//
//  D'tor
//--------------------------------------------------------------------------//
ModuleGuiBuilder::~ModuleGuiBuilder()
{
    delete TreeViewMgr;
}

//--------------------------------------------------------------------------//
//  Build the Parameters tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildParametersTab( const Module* mod,
                                      const StringStringMap_t& optMap )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, mod->GetNumParams(), 2 );

    const string& lang = mod->GetLanguage();

    ParamVector::const_iterator it = mod->BeginParams();
    ParamVector::const_iterator itend = mod->EndParams();
    for( int row=0; it != itend; ++row, ++it )
    {
        const Param* p = *it;
        string desc = p->GetDescription( lang );
        if ( desc.empty() )
            desc = p->GetDescription( "en" ); // use English as fallback

        // Check if a value is already assigned to this parameter
        string paramVal;
        StringStringMap_t::const_iterator mit = optMap.find( p->GetName() );
        if ( mit != optMap.end() )
            paramVal = (*mit).second;

        switch( p->GetType() )
        {
            case PARAM_BOOL: // create a checkbox
            {
                GtkWidget* checkbox = gtk_check_button_new();
                if ( paramVal == "1" )
                    gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(checkbox),
                                                  TRUE );
                gtk_widget_show( checkbox );
                addToTable( table, checkbox, row, 0 );
                makeLabel( table, desc.c_str(), row, 1 );
                WidgetMap[ p ] = checkbox;
                break;
            }

            case PARAM_ARRAY:  // create a spin box
            {
                int minVal = atoi( p->GetMin().c_str() );
                int maxVal = atoi( p->GetMax().c_str() );

                GtkAdjustment* adj = (GtkAdjustment *)
                      gtk_adjustment_new( minVal, minVal, maxVal, 1, 2.0, 1.0);
                GtkWidget* spin = gtk_spin_button_new( adj, 1.0, 0 );
                if ( !paramVal.empty() )
                    gtk_spin_button_set_value( GTK_SPIN_BUTTON(spin),
                                               atoi( paramVal.c_str() ) );

                gtk_widget_show( spin );
                addToTable( table, spin, row, 0 );
                makeLabel( table, desc.c_str(), row, 1 );
                WidgetMap[ p ] = spin;
                break;
            }

            case PARAM_INT:
            case PARAM_STRING: // create a text or combo entry
            {
                GtkWidget* entry = NULL;
                GtkWidget* parent = NULL;
                if ( p->GetValues().empty() )
                    parent = entry = gtk_entry_new();
                else
                {
                    parent = gtk_combo_new();
                    entry = GTK_COMBO(parent)->entry;
                    // TODO: populate list of values
                }
                WidgetMap[ p ] = entry;

                if ( !paramVal.empty() )
                    gtk_entry_set_text( GTK_ENTRY(entry), paramVal.c_str() );

                gtk_widget_set_size_request( parent, 50, -1 );
                gtk_widget_show( parent );
                addToTable( table, parent, row, 0 );
                makeLabel( table, desc.c_str(), row, 1 );
                break;
            }

            default:
                break;
        }
    }

    if ( mod->GetNumParams() == 0 )
        makeLabel( table, _("This module has no parameters"), 0, 0 );

    replaceTab( Notebook, 0, scroll );
}

//--------------------------------------------------------------------------//
//  Build the Setup tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildSetupTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, 1, 1 );

    // TODO build setup tab
    makeLabel( table, _("This feature is not implemented yet"), 0, 0 );
    replaceTab( Notebook, 1, scroll );
}

//--------------------------------------------------------------------------//
//  Build the Updates tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildUpdatesTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table;

    // Display XSA infos, if any
    const XsaData* xsaData = mod->GetXsaData();
    if ( !xsaData )
    {
        table = makeTable( scroll, 1, 1 );
        makeLabel( table, _("I'm sorry, new versions informations for this module are not available"), 0, 0 );
    }
    else
    {
        table = makeTable( scroll, 2, 1 );
        GtkWidget* label = makeLabel( table,
                                  _("The latest versions of this module are:"),
                                  0, 0 );
        gtk_misc_set_padding( GTK_MISC( label ), 0, 10 );

        // Display versions as a list
        GtkWidget* treeView = gtk_tree_view_new();
        addToTable( table, treeView, 1, 0 );
        gtk_widget_show( treeView );

        delete TreeViewMgr;
        TreeViewMgr = new XsaTreeViewManager( GTK_TREE_VIEW(treeView),
                                              xsaData );
        TreeViewMgr->InitListStore();
    }

    replaceTab( Notebook, 2, scroll );
}

//--------------------------------------------------------------------------//
//  Build the About tab in the notebook widget
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::buildAboutTab( const Module* mod )
{
    GtkWidget* scroll = makeScrolledWindow();
    GtkTable* table = makeTable( scroll, 4, 2 );

    int row = 0;
    makeLabel( table, _("Author:"), row, 0 );
    makeLabel( table, mod->GetAuthors(), row++, 1 );
    makeLabel( table, _("License:"), row, 0 );
    makeLabel( table, mod->GetLicense(), row++, 1 );

    if ( !mod->GetURL().empty() )
    {
        makeLabel( table, _("Website:"), row, 0 );
        GtkWidget* url = gnome_href_new( mod->GetURL().c_str(),
                                         mod->GetURL().c_str() );
        gtk_widget_show( url );
        addToTable( table, url, row++, 1 );
    }

    const string& lang = mod->GetLanguage();
    string about = mod->GetAbout( lang );
    if ( about.empty() )
        about = mod->GetAbout( "en" ); // use English as fallback

    if ( !about.empty() )
    {
        makeLabel( table, _("Description:"), row, 0 );
        makeLabel( table, about, row++, 1 );
    }

    replaceTab( Notebook, 3, scroll );
}

//--------------------------------------------------------------------------//
//  Update module name and usage at top of window
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::updateNameAndUsage( const Module* mod )
{
    GtkLabel* modname = GTK_LABEL( lookup_widget(TopWidget, "modname") );
    gtk_label_set_text( modname, mod->GetName().c_str() );

    const string& lang = mod->GetLanguage();
    string usageStr = mod->GetUsage( lang );
    if ( usageStr.empty() )
        usageStr = mod->GetUsage( "en" ); // use English as fallback

    GtkLabel* usage = GTK_LABEL( lookup_widget(TopWidget, "usage") );
    gtk_label_set_text( usage, usageStr.c_str() );
}

//--------------------------------------------------------------------------//
//  Dynamically fill in the notebook tabs with Module data
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::BuildGui( const Module* mod, const string& options )
{
    Notebook = GTK_NOTEBOOK( lookup_widget(TopWidget, "notebook") );
    CurrentModule = mod;

    // Forget the previous list of widgets
    WidgetMap.clear();

    // Parse the options string, and turn it into a form we can use
    StringStringMap_t optMap;
    parseOptions( options, optMap );

    if ( Verbosity )
        cout << "BuildGui: " << mod->GetName()
             << "'s options=\"" << options << "\"" << endl;

    buildParametersTab( mod, optMap );
    buildSetupTab( mod );
    buildUpdatesTab( mod );
    buildAboutTab( mod );
    updateNameAndUsage( mod );

    gtk_notebook_set_current_page( Notebook, 0 );
}

//--------------------------------------------------------------------------//
//  Parse an options string
//--------------------------------------------------------------------------//
void
ModuleGuiBuilder::parseOptions( const string& options,
                                StringStringMap_t& optMap ) const
{
    std::istringstream istr( options );
    std::istream_iterator<string> is( istr ), isend;
    for( ; is != isend; ++is )
    {
        const string& opt = *is;
        int equal = opt.find_first_of("=");
        string paramName = opt.substr( 0, equal );
        string paramVal = opt.substr( equal+1, opt.length() );
        optMap[ paramName ] = paramVal;
    }
}

//--------------------------------------------------------------------------//
//  Generate an option string for modules.conf
//--------------------------------------------------------------------------//
string
ModuleGuiBuilder::GenerateOptionsString() const
{
    std::ostringstream oss;
    MapParamWidget::const_iterator it = WidgetMap.begin();
    MapParamWidget::const_iterator itend = WidgetMap.end();
    for( ; it != itend; ++it )
    {
        const Param* param = (*it).first;
        const GtkWidget* widget = (*it).second;

        switch( param->GetType() )
        {
            case PARAM_BOOL:
            {
                if ( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(widget) ) )
                    oss << " " << param->GetName() << "=1";
                break;
            }
            case PARAM_ARRAY:
            {
                gint val = gtk_spin_button_get_value_as_int(
                                                    GTK_SPIN_BUTTON(widget) );
                // Default is the "min" value
                if ( val != atoi( param->GetMin().c_str() ) )
                    oss << " " << param->GetName() << "=" << val;
                break;
            }
            case PARAM_INT:
            case PARAM_STRING:
            {
                const gchar* text = gtk_entry_get_text( GTK_ENTRY(widget) );
                if ( strcmp( text, "" ) )
                    oss << " " << param->GetName() << "=" << text;
                break;
            }
        }
    }

    return oss.str();
}

