//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  modinforeader.cpp: gather module informations from XML and modinfo
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "modinforeader.h"
#include "moduletypes.h"
#include "pipereader.h"
#include <gnome.h>
#include <libgnomevfs/gnome-vfs.h>
#include <stdexcept>
#include <iostream>
#include <iterator>
#include <sstream>
#include <libxml++/libxml++.h>
#include "xsareader.h"

using std::cout;
using std::endl;
using std::ifstream;

using namespace gmodcfg;

//--------------------------------------------------------------------------//
//  D'tor
//--------------------------------------------------------------------------//
ModuleInfoReader::~ModuleInfoReader()
{
    ClearCache();
}

//--------------------------------------------------------------------------//
//  Helper function to parse a parameter definition in an XML tree
//  Returns: ptr to allocated Param object
//--------------------------------------------------------------------------//
static Param*
parseParam( const xmlpp::Node* node )
{
    const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(node);
    if ( !elem )
        return NULL;
    const xmlpp::Attribute* nameAttr = elem->get_attribute("name");
    const xmlpp::Attribute* typeAttr = elem->get_attribute("type");
    const xmlpp::Attribute* valuesAttr = elem->get_attribute("values");
    const xmlpp::Attribute* minAttr = elem->get_attribute("min");
    const xmlpp::Attribute* maxAttr = elem->get_attribute("max");

    string name, type;
    if ( nameAttr )
        name = nameAttr->get_value();
    if ( typeAttr )
        type = typeAttr->get_value();

    Param* p = new Param( name, type );

    if ( valuesAttr )
        p->SetValues( valuesAttr->get_value() );
    if ( minAttr )
        p->SetMin( minAttr->get_value() );
    if ( maxAttr )
        p->SetMax( maxAttr->get_value() );

    // Parse the description(s)
    const xmlpp::Node::NodeList& children = elem->get_children();
    xmlpp::Node::NodeList::const_iterator it = children.begin();
    xmlpp::Node::NodeList::const_iterator itend = children.end();
    for ( ; it != itend; ++it )
    {
        const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(*it);
        if ( !elem )
            continue;
        const xmlpp::Attribute* descLang = elem->get_attribute("lang");
        const xmlpp::TextNode* text = elem->get_child_content();
        if ( !text )
            continue;

        p->SetDescription( text->get_content(), descLang->get_value() );
    }

    return p;
}

//--------------------------------------------------------------------------//
//  Download an XSA file and parse it
//--------------------------------------------------------------------------//
static XsaData*
getXsaData( const string& src, int verbosity )
{
    char buf[] = "/tmp/gmodconfigXXXXXX";
    if ( mkstemp( buf ) == -1 )
        return 0;

    if ( verbosity )
        cout << "Getting XSA file from " << src << endl;

    GnomeVFSURI* srcUri = gnome_vfs_uri_new( src.c_str() );
    GnomeVFSURI* fileUri = gnome_vfs_uri_new( "file://" );
    GnomeVFSURI* destUri = gnome_vfs_uri_append_path( fileUri, buf );

    GnomeVFSResult result = gnome_vfs_xfer_uri( srcUri, destUri,
                            GNOME_VFS_XFER_DEFAULT,
                            GNOME_VFS_XFER_ERROR_MODE_ABORT,
                            GNOME_VFS_XFER_OVERWRITE_MODE_REPLACE,
                            NULL, NULL );
    if ( result != GNOME_VFS_OK )
    {
        cout << _("Error reading the XSA file: ")
             << gnome_vfs_result_to_string( result ) << endl;
        return 0;
    }

    std::ifstream ifs( buf );
    XsaReader reader( verbosity );
    XsaData* data = reader.Read( ifs );
    unlink( buf );

    return data;
}

//--------------------------------------------------------------------------//
//  Read in module information from an XML stream
//  Returns: ptr to an allocated Module object
//--------------------------------------------------------------------------//
static void
readXmlStream( istream& is, Module* mod, int verbosity )
{
    if ( verbosity )
        cout << "Reading XML from stream..." << endl;

    xmlpp::DomParser parser;
    if( !parser )
        throw std::runtime_error( _("Error while building the parser") );

    parser.parse_stream( is );

#if LIBXMLPP_MAJOR_VERSION==0 && LIBXMLPP_MINOR_VERSION<22
    const xmlpp::Node* root = parser.get_root_node(); //deleted by DomParser.
#else
    const xmlpp::Node* root = parser.get_document()->get_root_node(); //deleted by DomParser.
#endif

    if ( !root )
        throw std::runtime_error( _("Empty XML document") );

    // Traverse the document tree. First level is module
    const xmlpp::Element* modNode = dynamic_cast<const xmlpp::Element*>(root);

    const xmlpp::Attribute* modNameAttr = modNode->get_attribute("name");
    mod->SetName( modNameAttr->get_value() );
    if ( verbosity )
        cout << "module name=" << mod->GetName() << endl;

    // 2nd level is either param, or url, or shortdesc, or longdesc
    const xmlpp::Node::NodeList& children = modNode->get_children();
    xmlpp::Node::NodeList::const_iterator it = children.begin();
    xmlpp::Node::NodeList::const_iterator itend = children.end();
    for ( ; it != itend; ++it )
    {
        const xmlpp::Node* node = *it;

        if ( node->get_name() == "param" )
        {
            Param* p = parseParam( node );
            if ( p )
                mod->AddParam( p );
        }
        else if ( node->get_name() == "url" )
        {
            const xmlpp::Element* elem =
                                     dynamic_cast<const xmlpp::Element*>(*it);
            const xmlpp::TextNode* text = elem->get_child_content();
            if ( !text )
                continue;

            mod->SetURL( text->get_content() );
            if ( verbosity )
                cout << "url=" << mod->GetURL() << endl;
        }
        else if ( node->get_name() == "xsa" )
        {
            const xmlpp::Element* elem =
                                     dynamic_cast<const xmlpp::Element*>(*it);
            const xmlpp::TextNode* text = elem->get_child_content();
            if ( !text )
                continue;

            string xsaLoc = text->get_content();
            if ( verbosity )
                cout << "xsa=" << xsaLoc << endl;

            // Retrieve XSA file right now
            if ( !xsaLoc.empty() )
            {
                XsaData* data = getXsaData( xsaLoc, verbosity );
                mod->SetXsaURL( xsaLoc );
                mod->SetXsaData( data );
            }
        }
        else if ( node->get_name() == "usage" )
        {
            const xmlpp::Element* elem =
                                     dynamic_cast<const xmlpp::Element*>(*it);
            const xmlpp::Attribute* usageLang = elem->get_attribute("lang");
            string lang = usageLang->get_value();
            const xmlpp::TextNode* text = elem->get_child_content();
            if ( !text )
                continue;

            mod->SetUsage( text->get_content(), lang );
            if ( verbosity )
                cout << "usage (" << lang << ")=" << mod->GetUsage( lang )
                     << endl;
        }
        else if ( node->get_name() == "about" )
        {
            const xmlpp::Element* elem =
                                     dynamic_cast<const xmlpp::Element*>(*it);
            const xmlpp::Attribute* aboutLang = elem->get_attribute("lang");
            string lang = aboutLang->get_value();
            const xmlpp::TextNode* text = elem->get_child_content();
            if ( !text )
                continue;

            mod->SetAbout( text->get_content(), lang );
            if ( verbosity )
                cout << "about (" << lang << ")=" << mod->GetAbout( lang )
                     << endl;
        }
    }

    if ( verbosity )
        cout << "Done reading XML from stream." << endl;
}

//--------------------------------------------------------------------------//
//  Read in more informations from modinfo's output. This can also serve
//  as fallback in case no XML file is available for this module.
//  Returns: true if success
//--------------------------------------------------------------------------//
static bool
readFromModinfo( Module* mod, int verbosity=0 )
{
    if ( verbosity )
        cout << "Reading from modinfo..." << endl;

    if ( mod->GetName().empty() )
    {
        if ( verbosity )
            cout << "Module has no name, bailing out..." << endl;
        return false;
    }

    // FIXME: don't hardcode
    string cmd = "/sbin/modinfo " + mod->GetName();

    ifstream ifs;
    PipeReader pr( cmd, ifs );

    string line;
    while ( getline( ifs, line ) )
    {
        int colon = line.find_first_of(":");
        if ( colon == string::npos )  // should not happen
            continue;

        // Parameter descriptions start with "parm:"
        if ( line.compare( 0, colon, "parm" ) == 0 )
        {
            std::istringstream istr( line );
            std::istream_iterator<string> is( istr ), isend;
            ++is; // skip "parm:"

            string name = *is;

            // Perhaps no name is available...
            if ( name == "warning:" )
                continue;

            ++is;
            string type = *is;
            if ( type.find_first_of( ',' ) != string::npos )
                type = type.substr( 0, type.length()-1 );

            // Now move on to "array" or "description"
            ++is;

            string minStr, maxStr;
            if ( (is != isend) && (*is == "array") )
            {
                type = "array";
                ++is; ++is; ++is;
                minStr = string( *is, 0, (*is).length()-1 ); // remove ","
                ++is; ++is; ++is;
                maxStr = string( *is, 0, (*is).length()-2 ); // remove "),"
                ++is; // now we're on "description"
            }

            // Lookup this parameter. If it does not exist, create it
            Param* p = mod->Find( name );
            if ( !p )
            {
                p = new Param( name, type );
                mod->AddParam( p );
                p->SetMin( minStr );
                p->SetMax( maxStr );
            }

            bool hasDesc = ( is != isend );

            if ( verbosity )
            {
                cout << name << " [" << type << "] ";
                if ( type == "array" )
                    cout << "min=[" << minStr << "] max=[" << maxStr << "] ";
            }

            // We're overriding the parameter description (in English) only
            // if none are available already from the XML.
            if ( !p->GetDescription( "en" ).empty() )
                continue;

            // If no description available from modinfo, use module name
            if ( !hasDesc )
            {
                p->SetDescription( name, "en" );
                continue;
            }

            // Description is everything until the end of the line. We don't
            // need the istream_iterator anymore at this point.
            string desc = line.substr( istr.tellg(), line.length() );
            desc = desc.erase( desc.find_first_of( '"' ), 1 );
            int lastQuote = desc.find_last_of( '"' );
            if ( lastQuote != string::npos )
                desc = desc.erase( desc.find_last_of( '"' ), 1 );

            p->SetDescription( desc, "en" );

            if ( verbosity )
                cout << "desc=\"" << desc << "\"" << endl;
        }

        // Module's short description
        else if ( line.compare( 0, colon, "description" ) == 0 )
        {
            int descStart = line.find_first_of( '"', colon ) + 1;
            int descEnd = line.find_first_of( '"', descStart+1 );
            if ( mod->GetUsage().empty()  )
            {
                mod->SetUsage( line.substr( descStart, descEnd-descStart ),
                               "en" );
                if ( verbosity )
                    cout << "module description=\"" << mod->GetUsage()
                         << "\"" << endl;
            }
        }

        // Authors
        else if ( line.compare( 0, colon, "author" ) == 0 )
        {
            int authStart = line.find_first_of( '"', colon );
            if ( authStart == string::npos )
                authStart = colon;
            int authEnd = line.find_first_of( '"', authStart+1 );
            if ( authEnd == string::npos )
                authEnd = colon;
            else
                --authEnd; // don't include the end quote
            mod->SetAuthors( line.substr( authStart+1, authEnd-authStart ) );
        }

        // License
        else if ( line.compare( 0, colon, "license" ) == 0 )
        {
            int licStart = line.find_first_of( '"', colon ) + 1;
            int licEnd = line.find_first_of( '"', licStart+1 );
            mod->SetLicense( line.substr( licStart, licEnd-licStart ) );
        }
    }

    if ( verbosity )
        cout << "Done reading from modinfo." << endl;

    return true;
}

//--------------------------------------------------------------------------//
//  Main entry point
//  Assumption: a module whose name is "modName" exists
//  Returns: a new Module object if success, else NULL
//--------------------------------------------------------------------------//
const Module*
ModuleInfoReader::GetInfo( const string& modName, const string& lang )
{
    if ( Verbosity )
        cout << "\nGetting infos for module " << modName << "..." << endl;

    // Look up cache first
    Module* mod = getModule( modName );
    if ( mod )
    {
        if ( Verbosity )
            cout << "Found infos in cache" << endl;
        return mod;
    }

    // Not found in cache, need to create a new Module. First try reading
    // the associated XML file (if any) in central directory
    mod = new Module( modName );
    mod->SetLanguage( lang );

    string xmlFile = PACKAGE_DATA_DIR "/gmodconfig/xml/" + modName + ".xml";
    if ( Verbosity )
        cout << "Trying to access " << xmlFile << "..." << endl;

    ifstream ifs ( xmlFile.c_str() );
    if ( ifs.good() )
    {
        try
        {
            readXmlStream( ifs, mod, Verbosity );
        }
        catch( const std::exception& e )
        {
            if ( Verbosity )
                cout << "Exception caught in ModuleInfoReader: "
                     << e.what() << endl;
        }
    }

    // Parse modinfo's output to get more informations (or as fallback)
    try
    {
        readFromModinfo( mod, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader: "
                 << e.what() << endl;
    }

    // Update our cache
    addModule( modName, mod );

    return mod;
}

//--------------------------------------------------------------------------//
//  Read from modinfo's output
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::ReadFromModinfo( const string& modName ) const
{
    Module* mod = new Module( modName );

    bool ok = false;
    try
    {
        ok = readFromModinfo( mod, Verbosity );
    }
    catch( const std::exception& e )
    {
        ok = false;
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader::ReadFromModinfo "
                 << e.what() << endl;
    }

    if ( !ok )
    {
        delete mod;
        mod = 0;
    }

    return mod;
}

//--------------------------------------------------------------------------//
//  Read from a stream
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::ReadFromStream( istream& is ) const
{
    Module* mod = new Module("");

    try
    {
        readXmlStream( is, mod, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in ModuleInfoReader: "
                 << e.what() << endl;
    }
    return mod;
}

//--------------------------------------------------------------------------//
//  Module infos cache helpers
//--------------------------------------------------------------------------//
Module*
ModuleInfoReader::getModule( const string& str ) const
{
    StringModuleMap_t::const_iterator it = ModuleCache.find( str );
    if ( it != ModuleCache.end() )
        return (*it).second;
    return 0;
}

void
ModuleInfoReader::addModule( const string& str, Module* newmod )
{
    if ( getModule( str ) )
        return;

    if ( Verbosity )
        cout << "Adding infos for " << str << " module to cache" << endl;

    ModuleCache[ str ] = newmod;
}

void
ModuleInfoReader::replaceModule( const string& str, Module* newmod )
{
    StringModuleMap_t::iterator it = ModuleCache.find( str );
    if ( it != ModuleCache.end() )
    {
        delete (*it).second;
        (*it).second = newmod;
    }
}

//--------------------------------------------------------------------------//
//  Throw away the cache
//--------------------------------------------------------------------------//
void
ModuleInfoReader::ClearCache()
{
    StringModuleMap_t::iterator it = ModuleCache.begin();
    StringModuleMap_t::iterator itend = ModuleCache.end();
    for( ; it != itend; ++it )
        delete (*it).second;
    ModuleCache.clear();
}

