//  gmodconfig - a Gnome based kernel module configuration tool
//  http://gmodconfig.sourceforge.net/
//  Copyright (C) 2003 Cyril Bortolato
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  xsareader.cpp: read an XSA file from an istream
//--------------------------------------------------------------------------//

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "xsareader.h"
#include "xsatypes.h"
#include <gnome.h>
#include <stdexcept>
#include <iostream>
#include <fstream>
#include <iterator>
#include <sstream>
#include <libxml++/libxml++.h>

using std::cout;
using std::endl;
using std::istream;

using namespace gmodcfg;

//--------------------------------------------------------------------------//
//  Helper function to parse a vendor definition in an XSA's XML tree
//--------------------------------------------------------------------------//
static void
parseVendor( XsaData* xsaData, const xmlpp::Node* node )
{
    const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(node);
    if ( !elem )
        return;

    // Parse the child nodes.
    const xmlpp::Node::NodeList& children = elem->get_children();
    xmlpp::Node::NodeList::const_iterator it = children.begin();
    xmlpp::Node::NodeList::const_iterator itend = children.end();
    for ( ; it != itend; ++it )
    {
        const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(*it);
        if ( !elem )
            continue;

        const xmlpp::TextNode* text = elem->get_child_content();
        if ( !text )
            continue;

        if ( elem->get_name() == "name" )
            xsaData->SetVendorName( text->get_content() );
        else if ( elem->get_name() == "email" )
            xsaData->SetVendorEmail( text->get_content() );
        else if ( elem->get_name() == "url" )
            xsaData->SetVendorUrl( text->get_content() );
    }
}

//--------------------------------------------------------------------------//
//  Helper function to parse a product definition in an XSA's XML tree
//  Returns: ptr to allocated Product object
//--------------------------------------------------------------------------//
static void
parseProduct( XsaData* xsaData, const xmlpp::Node* node )
{
    const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(node);
    if ( !elem )
        return;
    const xmlpp::Attribute* idAttr = elem->get_attribute("id");

    string id, name, version, lastrel, url, changes;
    id = idAttr->get_value();

    // Parse the child nodes.
    const xmlpp::Node::NodeList& children = elem->get_children();
    xmlpp::Node::NodeList::const_iterator it = children.begin();
    xmlpp::Node::NodeList::const_iterator itend = children.end();
    for ( ; it != itend; ++it )
    {
        const xmlpp::Element* elem = dynamic_cast<const xmlpp::Element*>(*it);
        if ( !elem )
            continue;

        const xmlpp::TextNode* text = elem->get_child_content();
        if ( !text )
            continue;

        if ( elem->get_name() == "name" )
            name = text->get_content();
        else if ( elem->get_name() == "version" )
            version = text->get_content();
        else if ( elem->get_name() == "last-release" )
            lastrel = text->get_content();
        else if ( elem->get_name() == "info-url" )
            url = text->get_content();
        else if ( elem->get_name() == "changes" )
            changes = text->get_content();
    }

    xsaData->AddProduct( id, name, version, lastrel, url, changes );
}

//--------------------------------------------------------------------------//
//  Read in XSA from a stream
//  Returns: ptr to an allocated XsaData object
//--------------------------------------------------------------------------//
static XsaData*
readXsaFromStream( istream& is, int verbosity )
{
    xmlpp::DomParser parser;
    if( !parser )
        throw std::runtime_error( _("Error while building the parser") );

    parser.parse_stream( is );

#if LIBXMLPP_MAJOR_VERSION==0 && LIBXMLPP_MINOR_VERSION<22
    const xmlpp::Node* root = parser.get_root_node(); //deleted by DomParser.
#else
    const xmlpp::Node* root = parser.get_document()->get_root_node(); //deleted by DomParser.
#endif

    if ( !root )
        throw std::runtime_error( _("Empty XML document") );

    if ( verbosity )
        cout << "Reading XSA,,," << endl;

    // Traverse the document tree. First level is xsa
    const xmlpp::Element* modNode = dynamic_cast<const xmlpp::Element*>(root);

    XsaData* xsaData = new XsaData;

    // 2nd level is either vendor, or product
    const xmlpp::Node::NodeList& children = modNode->get_children();
    xmlpp::Node::NodeList::const_iterator it = children.begin();
    xmlpp::Node::NodeList::const_iterator itend = children.end();
    for ( ; it != itend; ++it )
    {
        const xmlpp::Node* node = *it;

        if ( node->get_name() == "vendor" )
        {
            parseVendor( xsaData, node );
        }
        else if ( node->get_name() == "product" )
        {
            parseProduct( xsaData, node );
        }
    }

    return xsaData;
}

//--------------------------------------------------------------------------//
//  Main entry point
//  Returns: a new XsaData object if success, else NULL
//--------------------------------------------------------------------------//
XsaData*
XsaReader::Read( istream& is ) const
{
    XsaData* xsaData = NULL;

    try
    {
        xsaData = readXsaFromStream( is, Verbosity );
    }
    catch( const std::exception& e )
    {
        if ( Verbosity )
            cout << "Exception caught in XsaReader: " << e.what() << endl;
    }

    if ( Verbosity )
        cout << "Done reading XSA" << endl;

    return xsaData;
}

